;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "org-vcard" "0.3.1"
  "Org-mode support for vCard export and import."
  '((emacs "24.4"))
  :url "https://github.com/pinoaffe/org-vcard"
  :commit "03c504c34e5c31091d971090b249064e332987d7"
  :revdesc "03c504c34e5c"
  :keywords '("outlines" "org" "vcard")
  :authors '(("Alexis" . "flexibeast@gmail.com")
             ("Will Dey" . "will123dey@gmail.com")
             ("pinoaffe" . "pinoaffe@gmail.com"))
  :maintainers '(("pinoaffe" . "pinoaffe@gmail.com")))
